#include <stdio.h>
#include <ftdi.h>
#include <stdlib.h>
#include <string.h>

#define IIC_EEPROM 208
#define IIC_RAM 210

int VFOhigh, P, num_bytes=0;
unsigned char Div1N, Xtal=155, FT_port, buf[65536];

struct bands
{
  int VFOhigh[6], Min[6], Max[6];
  unsigned char Div1N[6];
} bands_struct;

void RAM(int);
void I2C_Init(void);
void Start(void);
void Stop(void);
void SDA(int);
void SCL(int);
void Ausgeben(unsigned char);
double Freq(int);
void InitCal(unsigned char*, int);
void init_bands(void);
void calc_P(double);

int main(int argc, const char* argv[])
{
  struct ftdi_context ftdic;
  int f, i, band=-1, initcal_eeprom=0, initcal_ram=0, set_freq=0, write_pins=0;
  double freq, offset=15, x;
  FILE* InitCal_file;
  char filename[256];
  unsigned char InitCal_array[256];

  f=ftdi_init(&ftdic);
  if(f<0)
  {
    fprintf(stderr, "ftdi_init failed: %d (%s)\n", f, ftdi_get_error_string(&ftdic));
    return EXIT_FAILURE;
  }
  //printf("ftdi_init succeeded\n");
  
  f=ftdi_usb_open(&ftdic, 0x0403, 0x6001);
  if(f<0)
  {
    fprintf(stderr, "ftdi_usb_open failed: %d (%s)\n", f, ftdi_get_error_string(&ftdic));
    return EXIT_FAILURE;
  }
  //printf("ftdi_usb_open succeeded\n");
  
  f=ftdi_set_bitmode(&ftdic, 0xFF, BITMODE_BITBANG);
  if(f<0)
  {
    fprintf(stderr, "ftdi_set_bitmode failed: %d (%s)\n", f, ftdi_get_error_string(&ftdic));
    return EXIT_FAILURE;
  }
  //printf("ftdi_set_bitmode succeeded\n");

  f=ftdi_set_baudrate(&ftdic, 38400);
  if(f<0)
  {
    fprintf(stderr, "ftdi_set_baudrate failed: %d (%s)\n", f, ftdi_get_error_string(&ftdic));
    return EXIT_FAILURE;
  }
  //printf("ftdi_set_baudrate succeeded\n");

  f=ftdi_read_pins(&ftdic, &FT_port);
  if(f<0)
  {
    fprintf(stderr, "ftdi_read_pins failed: %d (%s)\n", f, ftdi_get_error_string(&ftdic));
    return EXIT_FAILURE;
  }

  init_bands();
  
  if(!(argc>1))
  {
    return EXIT_SUCCESS;
  }
  
  strcpy(filename, "InitCal.bin");

  for(i=1; i<argc; i++)
  {
    if(!strncmp(argv[i], "--initcal-eeprom", 16))
    {
      initcal_eeprom=1;
      continue;
    }
    if(!strncmp(argv[i], "--initcal-ram", 13))
    {
      initcal_ram=1;
      continue;
    }
    if(!strncmp(argv[i], "--initcal-file=", 15))
    {
      f=strlen(argv[i]+15);
      if(f<1 || f>256)
      {
	return EXIT_FAILURE;
      }
      else
      {
        strcpy(filename, argv[i]+15);
        continue;
      }
    }
    if(!strncmp(argv[i], "--band=", 7))
    {
      f=atoi(argv[i]+7);
      if(f<0 || f>5)
      {
	return EXIT_FAILURE;
      }
      else
      {
	band=f;
        continue;
      }
    }
    if(!strncmp(argv[i], "--freq=", 7))
    { 
      x=atof(argv[i]+7);
      /*if(x==HUGE_VAL || x==-HUGE_VAL)
      {
	return EXIT_FAILURE;
      }
      else
      {*/
	freq=x;
	set_freq=1;
        continue;
      //}
    }
    if(!strncmp(argv[i], "--offset=", 9))
    {
      x=atof(argv[i]+9);
      /*if(x==HUGE_VAL || x==-HUGE_VAL)
      {
	return EXIT_FAILURE;
      }
      else
      {*/
	offset=x;
        continue;
      //}
    }
    if(!strncmp(argv[i], "--xtal=", 7))
    {
      f=atoi(argv[i]+7);
      if(f<0 || f>255)
      {
	return EXIT_FAILURE;
      }
      else
      {
	Xtal=f;
        continue;
      }
    }
    if(!strncmp(argv[i], "--mux=", 6))
    {
      f=atoi(argv[i]+6);
      if(f<0 || f>7)
      {
	return EXIT_FAILURE;
      }
      else
      {
        FT_port=FT_port&0x63;
        FT_port=FT_port+4*f;
	write_pins=1;
        continue;
      }
    }
    if(!strncmp(argv[i], "--db=", 5))
    {
      f=atoi(argv[i]+5);
      if(f<0 || f>2)
      {
	return EXIT_FAILURE;
      }
      else
      {
        FT_port=FT_port&0x1F;
        FT_port=FT_port+32*f;
	write_pins=1;
        continue;
      }
    }
    return EXIT_FAILURE;
  }

  if(initcal_eeprom || initcal_ram)
  {
    InitCal_file=fopen(filename,"rb");
  
    if(InitCal_file==NULL)
    {
      printf("fopen failed\n");
      return EXIT_FAILURE;
    }
  
    if(!(fread(InitCal_array, sizeof(InitCal_array), 1, InitCal_file)==1 && EOF))
    {
      printf("Failure: Expected 256 bytes\n");
      return EXIT_FAILURE;
    }
  
    if(initcal_eeprom)
    {
      InitCal(InitCal_array, IIC_EEPROM);
    }
    
    if(initcal_ram)
    {
      InitCal(InitCal_array, IIC_RAM);
    }
  
    fclose(InitCal_file);
  }
  
  if(band!=-1)
  {
    VFOhigh=bands_struct.VFOhigh[band];
    Div1N=bands_struct.Div1N[band];
  
    if(set_freq)
    {
      calc_P(freq-offset);
  
      //printf("P: %d\n", P);
  
      if(P<bands_struct.Min[band] || P>bands_struct.Max[band])
      {
	printf("Fmin: %f kHz\n", Freq(bands_struct.Min[band]));
        printf("Fmax: %f kHz\n", Freq(bands_struct.Max[band]));
        return EXIT_FAILURE;
      }
    
      printf("Setting frequency: %f kHz\n", Freq(P));
  
      RAM(IIC_RAM);
    }
    else
    {
      printf("Fmin: %f kHz\n", Freq(bands_struct.Min[band]));
      printf("Fmax: %f kHz\n", Freq(bands_struct.Max[band]));
    }
  }

  if(!(num_bytes>0) && write_pins)
  {
    buf[num_bytes]=FT_port;
    num_bytes++;
  }
  
  //printf("Number of bytes that would be written: %d\n", num_bytes);
  
  /*printf("Byte\tSDA\tSCL\tC\tB\tA\t9\t10\n");
  for(i=0; i<num_bytes; i++)
  {
    printf("%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n", i+1, buf[i]&1, (buf[i]&2)/2, (buf[i]&4)/4, (buf[i]&8)/8, (buf[i]&16)/16, (buf[i]&32)/32, (buf[i]&64)/64);
  }*/
  
  f=ftdi_write_data_set_chunksize(&ftdic, 16384);
  if(f<0)
  {
    fprintf(stderr, "ftdi_write_data_set_chunksize failed: %d (%s)\n", f, ftdi_get_error_string(&ftdic));
    return EXIT_FAILURE;
  }
  //printf("ftdi_write_data_set_chunksize succeeded\n");

  f=ftdi_write_data(&ftdic, buf, num_bytes);
  if(f<0)
  {
    fprintf(stderr,"ftdi_write failed: %d (%s)\n",f, ftdi_get_error_string(&ftdic));
  }
  printf("Number of bytes written: %d\n", f);
  
  /*printf("Disabling bitbang mode...\n");
  ftdi_disable_bitbang(&ftdic);*/

  ftdi_usb_close(&ftdic);
  ftdi_deinit(&ftdic);
  
  return EXIT_SUCCESS;
}

void RAM(int IICadr)
{
  int Pump=1;
  unsigned char R40, R41, R42;
  
  if(P>479)
  {
    Pump=2;
  }
  if(P>639)
  {
    Pump=3;
  }
  if(P>799)
  {
    Pump=4;
  }
  
  R40=(P/2-4)/256+4*Pump+192;
  R41=(P/2-4)&255;
  if(VFOhigh)
  {
    R42=23+128*(P%2);
  }
  else
  {
    R42=48+128*(P%2);
  }
  
  /*printf("R40: %u\n", R40);
  printf("R41: %u\n", R41);
  printf("R42: %u\n", R42);*/
  
  I2C_Init();
  
  Start();
  Ausgeben(IICadr);
  Ausgeben(64);
  Ausgeben(R40);
  Ausgeben(R41);
  Ausgeben(R42);
  Stop();
  
  Start();
  Ausgeben(IICadr);
  Ausgeben(12);
  Ausgeben(Div1N);
  Stop();
  
  Start();
  Ausgeben(IICadr);
  Ausgeben(19);
  Ausgeben(Xtal);
  Stop();
}

void I2C_Init(void)
{
  SCL(1); SDA(1);
}
  
void Start(void)
{
  SDA(0); SCL(0);
}

void Stop(void)
{
  SCL(0); SDA(0);
  SCL(1); SDA(1);
}

void SDA(int TXD)
{
  if(TXD)
  {
    FT_port=FT_port|1;
  }
  else
  {
    FT_port=FT_port&(255-1);
  }
  buf[num_bytes]=FT_port;
  num_bytes++;
}

void SCL(int RXD)
{  
  if(RXD)
  {
    FT_port=FT_port|2;
  }
  else
  {
    FT_port=FT_port&(255-2);
  }
  buf[num_bytes]=FT_port;
  num_bytes++;
}

void Ausgeben(unsigned char Wert)
{
  int i=128;
  
  while(i>0)
  {
    if(Wert&i)
    {
      SDA(1);
    }
    else
    {
      SDA(0);
    }
    SCL(1);
    SCL(0);
    i=i/2;
  }
  SDA(1); SCL(1);
  SCL(0);
}

double Freq(int P)
{
  if(VFOhigh)
  {
    return (double)(P*400/Div1N)/4;
  }
  else
  {
    return (double)(P*200/Div1N)/4;
  }
}

void InitCal(unsigned char* InitCal_array, int IICadr)
{
  int i, k;
  
  I2C_Init();
  
  for(i=0;i<16;i++)
  {  
    Start();
    Ausgeben(IICadr);
    Ausgeben(i*16);
    for(k=0;k<16;k++)
    {
      Ausgeben(InitCal_array[i*16+k]);
      //Ausgeben(0);
    }
    Stop();
  }
}

void init_bands(void)
{
  bands_struct.VFOhigh[0]=1;
  bands_struct.VFOhigh[1]=0;
  bands_struct.VFOhigh[2]=0;
  bands_struct.VFOhigh[3]=0;
  bands_struct.VFOhigh[4]=0;
  bands_struct.VFOhigh[5]=0;
  
  bands_struct.Min[0]=400;
  bands_struct.Min[1]=400;
  bands_struct.Min[2]=400;
  bands_struct.Min[3]=400;
  bands_struct.Min[4]=400;
  bands_struct.Min[5]=260;
  
  bands_struct.Max[0]=1200;
  bands_struct.Max[1]=1600;
  bands_struct.Max[2]=1600;
  bands_struct.Max[3]=1600;
  bands_struct.Max[4]=1600;
  bands_struct.Max[5]=1600;
  
  bands_struct.Div1N[0]=4;
  bands_struct.Div1N[1]=5;
  bands_struct.Div1N[2]=10;
  bands_struct.Div1N[3]=25;
  bands_struct.Div1N[4]=50;
  bands_struct.Div1N[5]=100;
}

void calc_P(double freq)
{
  if(freq<130 || freq>30000)
  {
    P=0;
  }
  if(VFOhigh)
  {
    P=(int)(freq*4/(400/Div1N)+0.5);
  }
  else
  {
    P=(int)(freq*4/(200/Div1N)+0.5);
  }
}
